%{ 
  module R = Syntax
  module S = Tags.SymbolSet
  open Prog
%}

%token UNDERSCORE LANGLE RANGLE RBRACKET LBRACKET BAR COMMA HAT STAR EQUAL
%token LPAREN RPAREN EOF LBRACE RBRACE QUESTION PLUS SLASH COLON ARROW BANG
%token <string> IDENT
%token CHECK RUN

%right BAR
%right COMMA
%nonassoc STAR PLUS SLASH QUESTION

%start prog
%type <Prog.prog> prog
%%

prog:
 | phrase prog { $1 :: $2 }
 | EOF { [] }

phrase:
 | IDENT EQUAL syntax { Def ($1,$3) }
/* | LPAREN syntax RPAREN syntax { Infer ($2,$4) } */
 | CHECK syntax COLON syntax ARROW syntax { Check ($2, $4, $6) }
 | RUN syntax LPAREN syntax RPAREN { Run ($2, $4) } 

syntax:
 | syntax BAR syntax { R.Alt ($1,$3) }
 | syntax COMMA syntax { R.Seq ($1,$3) }
 | syntax STAR { R.Star $1 }
 | syntax PLUS { R.Plus $1 }
 | syntax QUESTION { R.Alt (R.Eps, $1) } 
 | tag LBRACKET syntax_or_nothing RBRACKET { R.Elem ($1,$3) } 
 | IDENT { R.Name $1 }
 | LPAREN syntax_or_nothing RPAREN { $2 }
 | LANGLE tag LBRACKET syntax_or_nothing RBRACKET QUESTION syntax BAR syntax RANGLE
     { R.Test ($2, $4, $7, $9) }
 | SLASH syntax { R.Sub $2 }

syntax_or_nothing:
 | syntax { $1 }
 | { R.Eps }

tag:
 | IDENT { S.finite [$1] }
 | HAT IDENT { S.cofinite [$2] }
 | LBRACE symbolset RBRACE { S.finite $2 }
 | HAT LBRACE symbolset RBRACE { S.cofinite $3 }
 | UNDERSCORE { S.any }

symbolset:
 | IDENT BAR symbolset { $1 :: $3 }
 | IDENT { [$1] }



