open Tags

type t =
  | Name of string
  | Seq of t * t
  | Alt of t * t
  | Star of t
  | Plus of t
  | Eps
  | Elem of SymbolSet.t * t
  | Sub of t
  | Test of SymbolSet.t * t * t * t

let defs = Hashtbl.create 1024

let register x (d:t) = 
  if Hashtbl.mem defs x then
    failwith ("Redefinition of the variable: " ^ x);
  Hashtbl.add defs x d

let def x =
  try Hashtbl.find defs x
  with Not_found ->
    failwith ("Variable undefined: " ^ x)


let rec print pri ppf = function
  | Name s -> Format.fprintf ppf "%s" s
  | Seq (r1,r2) -> 
      if pri >= 2 then Format.fprintf ppf "(";
      Format.fprintf ppf "%a,%a" (print 1) r1 (print 1) r2;
      if pri >= 2 then Format.fprintf ppf ")"
  | Alt (r1,r2) -> 
      if pri >= 1 then Format.fprintf ppf "(";
      Format.fprintf ppf "%a|%a" (print 0) r1 (print 0) r2;
      if pri >= 1 then Format.fprintf ppf ")"
  | Star r -> 
      Format.fprintf ppf "%a*" (print 2) r
  | Plus r -> 
      Format.fprintf ppf "%a+" (print 2) r
  | Eps ->
      Format.fprintf ppf "()"
  | Elem (sset,Eps) ->
      Format.fprintf ppf "%a[]" SymbolSet.print sset
  | Elem (sset,sub) ->
      Format.fprintf ppf "%a[%a]" SymbolSet.print sset (print 0) sub
  | Sub sub ->
      Format.fprintf ppf "/%a" (print 0) sub
  | Test (sset,sub,e1,e2) ->
      Format.fprintf ppf "<%a[%a] ? %a | %a>" 
	SymbolSet.print sset 
	(print 0) sub 
	(print 0) e1
	(print 0) e2

let print = print 0

let compare = compare
let equal = (=)
let hash = Hashtbl.hash


let to_string e =
  let b = Buffer.create 1024 in
  let ppf = Format.formatter_of_buffer b in
  print ppf e;
  Format.fprintf ppf "@.";
  Buffer.contents b
