open Tags
open Expr
open Syntax

let fold_left2 f accu l1 l2 =
  List.fold_left
    (fun accu x -> List.fold_left (fun accu y -> f accu x y) accu l2)
    accu l1

module Var = struct
  type t = Syntax.t * State.t * State.t
  let compare (e1,start1,end1) (e2,start2,end2) =
    let c = Syntax.compare e1 e2 in if c !=0 then c 
    else let c = State.compare start1 start2 in if c != 0 then c
    else State.compare end1 end2

  let is_same (_,qstart,qend) = State.equal qstart qend
end


module VarList = Slist.Make(Var)
module StateList = Slist.Make(State)

module V = struct
  type t = VarList.t * StateList.t * StateList.t

  let compare (vs1,pos1,neg1) (vs2,pos2,neg2) =
    let c = VarList.compare vs1 vs2 in if c != 0 then c
    else let c = StateList.compare pos1 pos2 in if c != 0 then c
    else StateList.compare neg1 neg2

  let inter (vs1,pos1,neg1) (vs2,pos2,neg2) =
    (VarList.union vs1 vs2,
     StateList.union pos1 pos2,
     StateList.union neg1 neg2)

  let any = [], [], []

  let var e qstart qend = [ e,qstart,qend ], [], []
  let state q = [], [q], []

  let minus_q (vs,pos,neg) q = (vs, pos, StateList.add q neg)
  let has_neg_q (_,_,neg) q = StateList.mem q neg
  let has_q (_,pos,_) q = StateList.mem q pos

  let triv_sub (vs1,pos1,neg1) (vs2,pos2,neg2) =
    VarList.subset vs2 vs1
    && StateList.subset pos2 pos1
    && StateList.subset neg2 neg1

(*
  let print ppf (vs,pos,neg) =
    let first = ref true in
    let sep () = if !first then first := false else Format.fprintf ppf "," in
    List.iter 
      (fun (e,qstart,qend) -> sep ();
	 Format.fprintf ppf "E%i(%i->%i)" e.Expr.id 
	   (State.id qstart) (State.id qend))
      vs;
    List.iter (fun q -> sep (); Format.fprintf ppf "%i" (State.id q)) pos;
    List.iter (fun q -> sep (); Format.fprintf ppf "-%i" (State.id q)) neg
*)
end

module VMap = Map.Make(V)

let memo = ref VMap.empty

module Node = struct
  type t = (SymbolSet.t * V.t) list

  let intersect =
    fold_left2
      (fun accu (s1,sub1) (s2,sub2) ->
	 let s = SymbolSet.inter s1 s2 in
	 if SymbolSet.is_empty s 
	 then accu 
	 else (s, V.inter sub1 sub2) :: accu
      )
end

let any = [ SymbolSet.any, V.any, V.any ]

let cons accu f snext =
  List.fold_left (fun accu (s,vsub) -> (s,vsub,snext)::accu) accu f
    
let intersect =
  fold_left2
    (fun accu (s1,sub1,next1) (s2,sub2,next2) ->
       let s = SymbolSet.inter s1 s2 in
       if SymbolSet.is_empty s 
       then accu 
       else (s, V.inter sub1 sub2, V.inter next1 next2) :: accu
    )
    []

(*
let rec descr e qstart qend accu sset pos neg =
  let accu' = descr_aux e qstart qend accu sset pos neg in
  Printf.eprintf "descr %i->%i => %i,%i\n" (State.id qstart) (State.id qend) (List.length accu) (List.length accu');
  accu'
*)
  
let rec descr e qstart qend accu sset pos neg =
  if SymbolSet.is_empty sset then accu
  else if State.equal qstart State.any 
    && State.equal qend State.any then ( sset, ([], pos, neg) ) :: accu
  else match e with
    | Eps -> 
	if State.equal qstart qend 
	then ( sset, ([], pos, neg) ) :: accu 
	else accu
    | Seq (e1,e2) ->
	List.fold_left
	  (fun accu q ->
	     Node.intersect accu 
	       (descr e1 qstart q [] sset pos neg) 
	       (descr e2 q qend [] sset pos neg)
	  )
	  accu
	  (State.between qstart qend)
    | Elem (symb, e) ->
	if SymbolSet.is_singleton symb then 
	  let symb = SymbolSet.sample symb in
	  List.fold_left
	    (fun accu (sset', qsub, qnext) ->
	       if State.equal qnext qend && SymbolSet.has sset' symb then
		 List.fold_left
		   (fun accu qfin -> descr e qsub qfin accu sset pos neg)
		   accu
		   (State.finals qsub)
	       else accu
	    )
	    accu
	    (State.trans qstart)
	else 
	  List.fold_left
	    (fun accu (sset', qsub, qnext) ->
	       if State.equal qnext qend then
		 List.fold_left
		   (fun accu qfin -> descr e qsub qfin accu 
		      (SymbolSet.inter sset sset') pos neg)
		   accu
		   (State.finals qsub)
	       else accu
	    )
	    accu
	    (State.trans qstart)
    | Test (sset', sub, ethen, eelse) ->
	let sub = Regexp.compile sub in
	let accu = 
	  if State.equal sub State.any then
	    descr ethen qstart qend accu (SymbolSet.inter sset sset') pos neg
	  else
	    let accu = 
	      if StateList.mem sub neg then accu
	      else descr ethen qstart qend accu (SymbolSet.inter sset sset') 
		(StateList.add sub pos) neg	  
	    in
	    let accu = 
	      if StateList.mem sub pos then accu
	      else
		descr eelse qstart qend accu sset 
		  pos (StateList.add sub neg) in
	    accu in
	let accu = 
	  descr eelse qstart qend accu (SymbolSet.diff sset sset') pos neg in
	accu
    | Sub eseq ->
	( sset, ([ eseq,qstart,qend ], pos, neg) ) :: accu
    | Name x ->
	descr (Syntax.def x) qstart qend accu sset pos neg
    | _ -> assert false


let descr_expr e qstart qend = 
  descr e qstart qend [] SymbolSet.any [] []


let descr_var (e,qstart,qend) =
  List.fold_left
    (fun accu q -> cons accu (descr_expr e qstart q) (V.var e q qend))
    []
    (State.between qstart qend)
  
let descr_q q =
  List.map
    (fun (sset,qsub,qnext) -> (sset,V.state qsub, V.state qnext))
    (State.trans q)

let minus_qs t qs =
  let rec aux0 accu sset sub next = function
    | [] -> (sset,sub,next) :: accu
    | q :: rest -> aux1 accu sset sub next rest (State.trans q)
  and aux1 accu sset sub next qs = function
    | [] -> aux0 accu sset sub next qs
    | (sset',qsub,qnext) :: rest ->
	let i = SymbolSet.inter sset sset' in
	if SymbolSet.is_empty i || V.has_neg_q sub qsub || V.has_neg_q next qnext
	then aux1 accu sset sub next qs rest
	else
	  let accu = 
	    if V.has_q sub qsub then accu
	    else aux1 accu i (V.minus_q sub qsub) next qs rest in
	  let accu = 
	    if V.has_q next qnext then accu
	    else aux1 accu i sub (V.minus_q next qnext) qs rest in
	  let d = SymbolSet.diff sset sset' in
	  if SymbolSet.is_empty d then accu
	  else aux1 accu d sub next qs rest
  in
  if qs == [] then t
  else
    List.fold_left
      (fun accu (sset,sub,next) -> aux0 accu sset sub next qs)
      []
      t

let triv_sub (sset1,sub1,next1) (sset2,sub2,next2) =
  SymbolSet.subset sset1 sset2 
  && V.triv_sub sub1 sub2
  && V.triv_sub next1 next2

let rec filter_triv_sub accu = function
  | [] -> accu
  | h::t ->
      if List.exists (triv_sub h) accu || List.exists (triv_sub h) t 
      then filter_triv_sub accu t
      else filter_triv_sub (h::accu) t

let rec descr_varset q (vs,pos,neg) =
  let t = List.fold_left (fun accu v -> intersect accu (descr_var v)) any vs in
  let t = List.fold_left (fun accu q -> intersect accu (descr_q q)) t pos in
  let t = minus_qs t neg in
  let t = filter_triv_sub [] t in
  let t = List.map (fun (s, l1, l2) -> (s, state_of l1, state_of l2)) t in

  let final =
    List.for_all Var.is_same vs
    && List.for_all State.final pos
    && not (List.exists State.final neg) in

  State.set q final t

and state_of (vs,pos,neg) =
  let vs = 
    List.filter (fun (_,qstart,qend) -> qstart != State.any || qend != State.any) vs in
  match vs,pos,neg with
  | [], [q], [] -> q
  | [], [], [] -> State.any
  | l ->
      try VMap.find l !memo
      with Not_found ->
	let q = State.mk () in
	memo := VMap.add l q !memo;
	descr_varset q l;
(*	Format.fprintf Format.std_formatter "State for %a: %i@." V.print l 
	  (State.id q); *)
	q

let infer e q =
  let f = State.finals q in
  let t = List.flatten (List.map (fun qend -> descr_var (e,q,qend)) f) in
  let t = filter_triv_sub [] t in
  let t = List.map (fun (s, l1, l2) -> (s, state_of l1, state_of l2)) t in

  let final = List.memq q f in
  State.cons final t
  
let diff q1 q2 =
  state_of ([], [q1], [q2])
